﻿using System;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.Globalization;
using System.Runtime.InteropServices;
using EnvDTE;
using Microsoft.VisualStudio.Shell;
using System.Windows.Forms;

namespace UIMapToolbox.VSPackage
{
    /// <summary>
    /// This is the class that implements the package exposed by this assembly.
    ///
    /// The minimum requirement for a class to be considered a valid package for Visual Studio
    /// is to implement the IVsPackage interface and register itself with the shell.
    /// This package uses the helper classes defined inside the Managed Package Framework (MPF)
    /// to do it: it derives from the Package class that provides the implementation of the 
    /// IVsPackage interface and uses the registration attributes defined in the framework to 
    /// register itself and its components with the shell.
    /// </summary>
    // This attribute tells the PkgDef creation utility (CreatePkgDef.exe) that this class is
    // a package.
    [PackageRegistration(UseManagedResourcesOnly = true)]
    // This attribute is used to register the informations needed to show the this package
    // in the Help/About dialog of Visual Studio.
    [InstalledProductRegistration("#110", "#112", "1.0", IconResourceID = 400)]
    // This attribute is needed to let the shell know that this package exposes some menus.
    [ProvideMenuResource("Menus.ctmenu", 1)]
    [Guid(GuidList.guidUIMapToolbox_VSPackagePkgString)]
    public sealed class UIMapToolbox_VSPackagePackage : Package
    {
        private UIMapToolbox.UI.MainForm _uiMapToolboxForm;

        /// <summary>
        /// Default constructor of the package.
        /// Inside this method you can place any initialization code that does not require 
        /// any Visual Studio service because at this point the package object is created but 
        /// not sited yet inside Visual Studio environment. The place to do all the other 
        /// initialization is the Initialize method.
        /// </summary>
        public UIMapToolbox_VSPackagePackage()
        {
            Trace.WriteLine(string.Format(CultureInfo.CurrentCulture, "Entering constructor for: {0}", this.ToString()));
        }

        /////////////////////////////////////////////////////////////////////////////
        // Overriden Package Implementation
        #region Package Members

        /// <summary>
        /// Initialization of the package; this method is called right after the package is sited, so this is the place
        /// where you can put all the initilaization code that rely on services provided by VisualStudio.
        /// </summary>
        protected override void Initialize()
        {
            Trace.WriteLine (string.Format(CultureInfo.CurrentCulture, "Entering Initialize() of: {0}", this.ToString()));
            base.Initialize();

            // Add our command handlers for menu (commands must exist in the .vsct file)
            OleMenuCommandService mcs = GetService(typeof(IMenuCommandService)) as OleMenuCommandService;
            if ( null != mcs )
            {
                // Create the command for the menu item.
                CommandID menuCommandID = new CommandID(GuidList.guidUIMapToolbox_VSPackageCmdSet, (int)PkgCmdIDList.cmdidOpenInUIMapToolbox);
                MenuCommand menuItem = new MenuCommand(MenuItemCallback, menuCommandID );
                mcs.AddCommand( menuItem );
            }
        }
        #endregion

        /// <summary>
        /// This function is the callback used to execute a command when the a menu item is clicked.
        /// See the Initialize method to see how the menu item is associated to this function using
        /// the OleMenuCommandService service and the MenuCommand class.
        /// </summary>
        private void MenuItemCallback(object sender, EventArgs e)
        {
            const string Title = "Open in UIMap Toolbox";

            DTE dte = this.GetService(typeof(DTE)) as DTE;
            if (dte == null)
            {
                MessageBox.Show("Unable to get DTE instance.", Title, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                return;
            }

            // Get selected file
            if (dte.SelectedItems.Count == 0)
            {
                MessageBox.Show("No items selected.", Title, MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            SelectedItem item = dte.SelectedItems.Item(1);
            ProjectItem pi = item.ProjectItem;
            if ((pi == null) || (pi.FileCount == 0))
            {
                MessageBox.Show("No filenames in selected item. Cannot continue.", Title, MessageBoxButtons.OK, MessageBoxIcon.Information);
                return;
            }

            string fileName = pi.FileNames[1];

            // Open it in UIMapToolbox window
            UIMapToolbox.UI.MainForm frm = null;
            int uiMapPane = 0; // left side

            try
            {
                if ((_uiMapToolboxForm == null) || (_uiMapToolboxForm.IsDisposed))
                {
                    // first instance, just load it
                    _uiMapToolboxForm = new UI.MainForm();
                    frm = _uiMapToolboxForm;
                }
                else
                {
                    // ask user if it should be opened in existing instance
                    string msg = "Open file in existing UIMap Toolbox instance (will be opened in right side)?";
                    DialogResult answer = MessageBox.Show(msg, Title, MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                    if (answer == DialogResult.Yes)
                    {
                        frm = _uiMapToolboxForm;
                        uiMapPane = 1; // right side
                    }
                    else
                    {
                        frm = new UI.MainForm();
                    }
                }

                frm.LoadFile(uiMapPane, fileName);
                frm.Show();
                frm.BringToFront();
            }
            catch (Exception ex)
            {
                string msg = String.Format("An error occured while loading UIMap. Please verify that it's a valid UIMap file.\n\nError: {0}", ex.Message);
                MessageBox.Show(msg, Title, MessageBoxButtons.OK, MessageBoxIcon.Asterisk);

                if (frm != null)
                    frm.Dispose();
            }
        }
    }
}